import os
import json
import argparse

from text_rank.summarize import summarize
from text_rank.keywords import keywords


def generate_output_name(input_name, task):
    path, f = os.path.split(input_name)
    f, ext = os.path.splitext(f)
    return os.path.join(path, f"{f}-{task}{ext}")


def main():
    parser = argparse.ArgumentParser("Run Text Rank")
    parser.add_argument("files", nargs="+", help="A list of files that contain documents to process.")
    parser.add_argument(
        "--task",
        default="summarize",
        choices={"summarize", "keywords"},
        help="Summarize the document or extract keywords?",
    )
    parser.add_argument(
        "--count", default=5, type=int, help="The number of sentences/keywords to return",
    )
    parser.add_argument(
        "--iters", "-i", type=int, default=20, help="The maximum number of iterations to run text rank for.",
    )
    parser.add_argument(
        "--convergence",
        "-c",
        type=float,
        default=0.0001,
        help="The threshold of differences between iterations that causes the algorithm to stop.",
    )
    parser.add_argument(
        "--verbose", "-v", action="store_true", help="Show progress bars when fitting text rank.",
    )
    parser.add_argument("--indent", default=2, type=int, help="Indent formatting for the output json")
    args = parser.parse_args()

    process = summarize if args.task == "summarize" else keywords
    file_ext = "summary" if args.task == "summarize" else "keywords"

    for file_name in args.files:
        with open(file_name) as f:
            document = json.load(f)
        if not isinstance(document, list):
            raise ValueError(f"Data should be a list, found: {type(document)}")
        if args.task == "summarize" and not all(isinstance(sent, str) for sent in document):
            raise ValueError(f"Data should be a list of strings")
        if args.task == "keywords" and not all(isinstance(token, dict) for token in document):
            raise ValueError(f"Data should be a list of dictionaries")

        result = list(process(document, args.count))

        output = generate_output_name(file_name, file_ext)
        with open(output, "w") as wf:
            json.dump(result, wf, indent=2)


if __name__ == "__main__":
    main()



import math
import random
from enum import Enum
from operator import itemgetter
from functools import singledispatch
from typing import List, Tuple, Dict, Optional
import numpy as np
from text_rank.graph import Graph, AdjacencyList, AdjacencyMatrix, Vertex


ConvergenceType = Enum("ConvergenceType", "ALL ANY")


def sum_edges(edges: Dict[str, float]) -> float:
    """Calculate the total weight for a collection of edges.

    These edges are represented as a target -> weight mapping and is based on the representation
    we use in the vertex object, the edge doesn't have any information about the source node

    :param edges: A map of edges where the key is the target and the value is the weight
    :returns: The total weight of all edges
    """
    return sum(edges.values())


def accumulate_score(vertex: Vertex, ws: List[float], denom: List[float]):
    """Accumulate the scores from all nodes that have incoming connections to you.

    :param vertex: The node we all looking at
    :param ws: The scores of each node in the graph
    :param denom: The precomputed sum of outgoing edges for each node in the graph
    :returns: The sum of incoming scores weighted by incoming edge strength normalized by outgoing strengths
    """
    acc = []
    for edge, weight in vertex.edges_in.items():
        acc.append(weight / denom[edge] * ws[edge])
    return math.fsum(acc)


@singledispatch
def text_rank_init(graph: Graph, seed: Optional[int] = None) -> Tuple[List[float], List[float]]:
    """Generate the initial scores for each node and pre-compute the outgoing strength.

    The sum of the weights for outbound edges for a given node doesn't change as text rank runs because
    it is based only on the values in the graph, not on ws for the node so we can pre-compute and reuse
    it instead of always recalculating it.

    :param graph: The graph we will run text rank on
    :param seed: A seed for the RNG if we want reproduceability
    :returns: The initial scores for each node and the edge normalization factor for each node
    """
    raise NotImplementedError


@text_rank_init.register(AdjacencyList)
def text_rank_init_list(graph: AdjacencyList, seed: Optional[int] = None) -> Tuple[List[float], List[float]]:
    denom: List[float] = []
    ws: List[float] = []
    random.seed(seed)
    for v in graph.vertices:
        total = sum_edges(v.edges_out)
        # If the sum off all outgoing edges of V_j is 0.0 then the incoming edge from V_j to V_i will be 0.0
        # We can use anything as the denominator and the value will still be zero
        denom.append(total if total != 0.0 else 1.0)
        ws.append(random.random())
    return ws, denom


@text_rank_init.register(AdjacencyMatrix)
def text_rank_init_matrix(graph: AdjacencyMatrix, seed: Optional[int] = None) -> Tuple[np.ndarray, np.ndarray]:
    np.random.seed(seed)
    ws = np.random.rand(graph.vertex_count)
    denom = np.reshape(np.sum(graph.adjacency_matrix, axis=1), (-1, 1))
    # If the sum off all outgoing edges of V_j is 0.0 then the incoming edge from V_j to V_i will be 0.0
    # We can use anything as the denominator and the value will still be zero
    denom[denom == 0.0] = 1.0
    return ws, denom


@singledispatch
def text_rank_update(graph: Graph, ws: List[float], denom: List[float], dampening: float = 0.85) -> List[float]:
    """Calculate new score for each node.

    This is the main step in text rank

    :param graph: The text rank graph
    :param ws: The scores for each node
    :param denom: The outbound weight normalization factor for each node, pre-computed for efficiency
    :param dampening: A scalar between 0 and 1. Used to simulate randomly jumping from one vertex to another.

    :returns: The updated scores for each node
    """
    raise NotImplementedError


@text_rank_update.register(AdjacencyList)
def text_rank_update_list(
    graph: AdjacencyList, ws: List[float], denom: List[float], dampening: float = 0.85
) -> List[float]:
    updates: List[float] = []
    for v in graph.vertices:
        acc = accumulate_scores(v, ws, denom)
        updates.append(acc)
    # We collect the updated scores for each node and apply them after. If we were to apply these
    # updates as they happen we would get different results than from the vectorized version used
    # in the adjacency matrix version
    for i, update in enumerate(updates):
        ws[i] = (1 - dampening) + dampening * update
    return ws


@text_rank_update.register(AdjacencyMatrix)
def text_rank_update_matrix(
    graph: AdjacencyMatrix, ws: np.ndarray, denom: np.ndarray, dampening: float = 0.85
) -> np.ndarray:
    update = np.dot(ws, graph.adjacency_matrix / denom)
    ws = (1 - dampening) + dampening * update
    return ws


@singledispatch
def text_rank_output(graph: Graph, ws: List[float]) -> List[Tuple[str, float]]:
    raise NotImplementedError


@text_rank_output.register(AdjacencyList)
def text_rank_output_list(graph: AdjacencyList, ws: List[float]) -> List[Tuple[str, float]]:
    return sorted(zip(map(lambda v: v.value, graph.vertices), ws), key=itemgetter(1), reverse=True)


@text_rank_output.register(AdjacencyMatrix)
def text_rank_output_matrix(graph: AdjacencyMatrix, ws: np.ndarray) -> List[Tuple[str, float]]:
    ws = np.reshape(ws, (-1,))
    return sorted(zip(graph.label2idx.keys(), ws), key=itemgetter(1), reverse=True)


def text_rank(
    graph: Graph,
    dampening: float = 0.85,
    convergence: float = 0.0001,
    convergence_type: ConvergenceType = ConvergenceType.ALL,
    niter: int = 200,
    seed: Optional[int] = None,
) -> List[Tuple[str, float]]:
    """Implementation of text rank from here https://www.aclweb.org/anthology/W04-3252.pdf

    :param graph: The graph we are running text rank on
    :param dampening: A scalar between 0 and 1. Used to simulate randomly jumping from one vertex to another.
    :param convergence: An early stopping criteria, when any or all of the node scores change by less than `convergence`
        we stop updating the graph. Set to `0` to turn off early stopping.
    :param convergence_type: Should we stop when all nodes move less than `convergence` or when a single node does
    :param niter: An upper bound on the number of iterations to run
    :param seed: A reproducability seed to initialization of the node scores.

    :returns: Pairs of (node label, scores) sorted by score
    """
    if not 0 <= dampening <= 1:
        raise ValueError(f"dampening must be between `0` and `1`, got {dampening}")
    converge = all if convergence_type is ConvergenceType.ALL else any

    ws_prev, denom = text_rank_init(graph, seed)

    for _ in range(niter):
        ws = text_rank_update(graph, ws_prev, denom, dampening)
        if converge(abs(p - c) < convergence for p, c in zip(ws_prev, ws)):
            break
        ws_prev = ws

    return text_rank_output(graph, ws)



from typing import Optional, Callable, Type, List
from text_rank.utils import overlap, norm_sentence
from text_rank.text_rank import text_rank, ConvergenceType
from text_rank.graph import sentence_graph, AdjacencyMatrix, Graph


def summarize(
    sentences: List[str],
    nsents: Optional[int] = None,
    keep_order: bool = True,
    dampening: float = 0.85,
    convergence: float = 0.0001,
    convergence_type: ConvergenceType = ConvergenceType.ALL,
    niter: int = 200,
    seed: Optional[int] = None,
    sim: Callable[..., float] = overlap,
    norm: Callable[[str], str] = norm_sentence,
    GraphType: Type[Graph] = AdjacencyMatrix,
) -> List[str]:
    """Summarize text.

    :param sentences: The sentences to summarize.
    :param nsents: The number of sentences to use in the summary, If `None` it uses the number of
        sentences to summarize divided by 3
    :param keep_order: Should summary sentences appear in the same order they appear in the in the
        text, if False sentences or ordered by their text rank score.
    :param dampening: A scalar between 0 and 1. Used to simulate randomly jumping from one vertex to another.
    :param convergence: An early stopping criteria, when any or all of the node scores change by less than `convergence`
        we stop updating the graph. Set to `0` to turn off early stopping.
    :param convergence_type: Should we stop when all nodes move less than `convergence` or when a single node does
    :param niter: An upper bound on the number of iterations to run
    :param seed: A reproducability seed to initialization of the node scores.
    :param sim: A callable that returns the similarity between two vertices, used to set the weight of the edge.
        The callable should have a signature like:
            sim(
                normed_s1,
                normed_s2,
                raw_s1=raw_s1,
                raw_s2=raw_s2,
                s1_idx=s1_idx,
                s2_idx=s2_idx,
            ) -> float:
        Where normed_s1/2 is the normalized strings of the two sentences, raw_s1/2 is the version of the sentence
        before getting normalized and s1/2_idx is the index of the sentences in the token list. This should
        facilitate both simple and complex similarity functions and also experiments that the actual flow of text
        to determine connections.
    :param norm: A function the returns a normalized version of the input sentence. Default implementation lowercases
        string and removes non alpha-numeric characters.
        This is used so simple similarity functions like the set overlap in the paper work well.
    :param GraphType: The Graph class to use.

    :returns: A list of sentences summarizing the original text.
    """
    graph, offsets = sentence_graph(sentences, sim, norm, GraphType)
    if nsents is None:
        nsents = len(sentences) // 3
    selected = text_rank(
        graph, dampening=dampening, convergence=convergence, convergence_type=convergence_type, niter=niter, seed=seed,
    )[:nsents]
    indices = [offsets[s[0]][0] for s in selected]
    if keep_order:
        return [sentences[i] for i in sorted(indices)]
    return [sentences[i] for i in indices]



from itertools import chain
from operator import itemgetter
from typing import Optional, Callable, Type, List, Dict, Set
from text_rank.utils import filter_pos, norm_token
from text_rank.text_rank import text_rank, ConvergenceType
from text_rank.graph import keyword_graph, AdjacencyMatrix, Graph


def join_adjacent_keywords(candidates: List[str], offsets: Dict[str, List[str]]) -> Set[str]:
    """Combine keywords that are adjacent in the source text into a single keyword.

    :param candidates: The keywords selected
    :param offsets: The offsets of the keywords into the original text
    :returns: The keywords created by joining adjacent ones
    """
    candidates = list(chain(*([(candidate, o) for o in offsets[candidate]] for candidate in candidates)))
    candidates = sorted(candidates, key=itemgetter(1))
    keywords = []
    current = []
    prev = None
    for kw, i in candidates:
        if not current or i == prev + 1:
            current.append(kw)
        else:
            keywords.append(" ".join(current))
            current = [kw]
        prev = i
    if current:
        keywords.append(" ".join(current))
    return set(keywords)


def keywords(
    tokens: List[Dict[str, str]],
    nwords: Optional[int] = None,
    winsz: int = 2,
    dampening: float = 0.85,
    convergence: float = 0.0001,
    convergence_type: ConvergenceType = ConvergenceType.ANY,
    niter: int = 200,
    seed: Optional[int] = None,
    sim: Callable[..., float] = lambda x, y, **kwargs: 1,
    norm: Callable[[str], str] = norm_token,
    filt: Callable[[Dict[str, str]], bool] = filter_pos,
    GraphType: Type[Graph] = AdjacencyMatrix,
) -> Set[str]:
    """Find keywords.

    :param tokens: The tokens of the text to pull keywords from
    :param nwords: The number of keywords to extract from the text. If `None` it uses the number
        of tokens in the passage divided by 3. This number isn't exact, after finding keywords it will
        join adjacent ones so it might be possible to get fewer keywords back.
    :param winsz: The size of a window around each word that it can connect to. This window is calculated
        in the non-filtered token list.
    :param dampening: A scalar between 0 and 1. Used to simulate randomly jumping from one vertex to another.
    :param convergence: An early stopping criteria, when any or all of the node scores change by less than `convergence`
        we stop updating the graph. Set to `0` to turn off early stopping.
    :param convergence_type: Should we stop when all nodes move less than `convergence` or when a single node does
    :param niter: An upper bound on the number of iterations to run
    :param seed: A reproducability seed to initialization of the node scores.
    :param sim: A callable that returns the similarity between two vertices, used to set the weight of the edge.
        The callable should have a signature like:
            sim(
                normed_s1,
                normed_s2,
                context=context,
                raw_s1=raw_s1,
                raw_s2=raw_s2,
                raw_context=raw_context,
                s1_idx=s1_idx,
                s2_idx=s2_idx,
            ) -> float:
        Where normed_s1/2 is the normalized string of two keywords, context is all of the normalized tokens,
        raw_s1/2/context are the prenormalized versions and s1/2_idx are the indices of the keywords in the
        original sentence. This should allow complex similarity functions. The context is the whole list of
        tokens, not just the window around one.
    :param norm: A function the returns a normalized version of the input string. Default implementation
        lowercases string and removes non alpha-numeric characters.
        This is used to unify similar vertices, i.e. `Hurricane` and `hurricane` should be the same vertex
        and will be with normalization.
    :param filt: A function that will filter tokens based on pos tags if the inputs are Dicts.
    :param GraphType: The Graph class to use.

    :returns: The keywords from the passage
    """
    graph, offsets = keyword_graph(tokens, winsz, sim, norm, filt, GraphType)
    if nwords is None:
        nwords = len(tokens) // 3
    keywords = text_rank(
        graph, dampening=dampening, convergence=convergence, convergence_type=convergence_type, niter=niter, seed=seed
    )[:nwords]
    keywords = join_adjacent_keywords([kw[0] for kw in keywords], offsets)
    return keywords



__version__ = "0.3.4"

# This is a new name to reference the text_rank as a modules instead of the function.
# This is used for mocking when testing while still allowing for `from text_rank import text_rank`
# by and end user
import text_rank.text_rank as text_rank_module
from text_rank.text_rank import text_rank
from text_rank.keywords import keywords
from text_rank.summarize import summarize
from text_rank.graph import (
    sentence_graph,
    keyword_graph,
    AdjacencyList,
    AdjacencyMatrix,
)

from pathlib import Path

DEMO_LOC = Path(__file__).parent / "data"



import re
from math import log
from typing import Any, Dict, List
from collections import defaultdict


ASCII = re.compile(r"[^a-z0-9]")
POS = re.compile(r"^[NJ]")


def filter_pos(token: Dict[str, str]) -> bool:
    """Should this token be included based on POS tags.

    This says yes to tokens that have pos tags of Adjective, Cardinal Number, any kind of
    noun and any kind of verb.

    :param token: A dict representation of the token with a pos tag in the `pos` key

    :returns: True if the token should be used, False otherwise
    """
    if not POS.match(token["pos"]) and token["pos"] != "ADJ" and token["pos"] != "CD":
        return False
    return True


def cooccurrence(s1: Any, s2: Any, **kwargs) -> int:
    """A similarity function used for keyword graph generation.

    Because the default keyword graphs use cooccurrence instead of weighting the edges
    so we always return 1.

    :returns: 1
    """
    return 1


def overlap(s1: str, s2: str, **kwargs) -> float:
    """A similarity function used for sentences.

    This is the similarity function defined in (Mihalcea and Tarau 2004)

    ```math
        Similarity(S_i, S_j) = \frac{|{w_k|w_k \in S_i & \w_k \in S_j}|}{\log(|S_i|) + \log(|S_j|)}
    ```

    Note:
        There is a weird edge case in this function when both sets have a single word. This results
        in a denominator of 0.0 which causes an error. I am currently forcing the denominator to
        `0.9256056207721526` in this case. This makes since because then both of them are a single
        word, there are two cases, one where the overlap is `0`, this results in a weight of `0` which
        makes sense. The other case is when they match. This results in a `1.0803737332167307` which
        was calculated by fitting this function and finding the value it predicts for `1`. See
        scripts/find-sim-of-one.py for this fitting.

    :param s1: The first item
    :param s2: The second item
    :returns: The similarity between s1 and s2
    """
    s1 = set(s1.split())
    s2 = set(s2.split())
    intersection = len(s1 & s2)
    norm = log(len(s1)) + log(len(s2))
    norm = 0.9256056207721526 if norm == 0.0 else norm
    return intersection / norm


def norm_sentence(sent: str) -> str:
    """Normalize a sentence by lower casing and removing punctuation from each word.

    :param sent: The white spaced joined sentence.
    :returns: The sentence joined with whitespace.
    """
    tokens = sent.split()
    tokens = [norm_token(token) for token in tokens]
    return " ".join(tokens)


def norm_token(token: str) -> str:
    """Normalize a token by lowercasing and removing non-alphanumeric characters.

    :param token: The word to be normalized
    :returns: The word lowercased and punctuation removed.
    """
    token = token.lower()
    return ASCII.sub("", token)


def build_vocab(tokens: List[str]) -> Dict[str, int]:
    """Build a vocab mapping tokens to indices.

    Produced indices are contiguous and a token has the index of where it first appeared in the tokens list.

    :param tokens: The list of tokens to map
    :returns: A mapping from token to index
    """
    vocab = defaultdict(lambda: len(vocab))
    for token in tokens:
        vocab[token]
    return {k: i for k, i in vocab.items()}



import re
from setuptools import setup, find_packages


def get_version(project_name):
    regex = re.compile(r"""^__version__ = ["'](\d+\.\d+\.\d+)["']$""")
    with open(f"{project_name}/__init__.py") as f:
        for line in f:
            m = regex.match(line)
            if m is not None:
                return m.groups(1)[0]


class About(object):
    NAME = "text_rank"
    VERSION = get_version(NAME)
    AUTHOR = "blester125"
    EMAIL = f"{AUTHOR}@gmail.com"
    URL = f"https://github.com/{AUTHOR}/{NAME}"
    DL_URL = f"{URL}/archive/{VERSION}.tar.gz"
    LICENSE = "MIT"
    DESCRIPTION = "Text Rank"


ext_modules = []

setup(
    name=About.NAME,
    version=About.VERSION,
    description=About.DESCRIPTION,
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author=About.AUTHOR,
    author_email=About.EMAIL,
    url=About.URL,
    download_url=About.DL_URL,
    license=About.LICENSE,
    python_requires=">=3.6",
    packages=find_packages(),
    package_data={
        "text_rank": [
            "text_rank/data/automatic-summarization-sents.json",
            "text_rank/data/automatic-summarization-tokens.json",
            "text_rank/data/paper-example-keywords.json",
            "text_rank/data/paper-example-summarize.json",
        ],
    },
    include_package_data=True,
    install_requires=["numpy",],
    setup_requires=[],
    extras_require={"test": ["pytest"],},
    keywords=["NLP"],
    entry_points={"console_scripts": ["text-rank-demo = text_rank.demo:main", "text-rank = text_rank.main:main",],},
    ext_modules=ext_modules,
    classifiers={
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering",
    },
)



from itertools import combinations
from collections import defaultdict
from typing import Optional, Dict, Union, Callable, Tuple, Type, List
import numpy as np
from text_rank.utils import overlap, cooccurrence, norm_sentence, norm_token, filter_pos, build_vocab


class Vertex:
    def __init__(self, value: str):
        """A vertex in a graph.

        :param value: The label for this vertex
        """
        self.value = value
        self._edges_out: Dict[int, float] = {}
        self._edges_in: Dict[int, float] = {}

    @property
    def edges_out(self) -> Dict[int, float]:
        """A mapping of target vertex to weight representing the edges with this vertex as the source."""
        return self._edges_out

    @property
    def edges_in(self) -> Dict[int, float]:
        """A mapping of source vertex to weight representing the edges that end at this vertex."""
        return self._edges_in

    @property
    def degree_in(self) -> int:
        """The number of edges that end at this vertex."""
        return len(self.edges_in)

    @property
    def degree_out(self) -> int:
        """The number of edges that start at this vertex."""
        return len(self.edges_out)

    def __str__(self) -> str:
        """A summary of this vertex."""
        return f"V(term={self.value}, in={self.degree_in}, out={self.degree_out})"


class Graph:
    def __init__(self, vertices: Union[Dict[str, int], List[str]]):
        """A directed simple graph.

        :param vertices: A mapping of vertex labels to integer indices or a list of vertex labels.
            If the latter then indices are assigned in order
        """
        if isinstance(vertices, dict):
            self.label2idx: Dict[str, int] = vertices
        else:
            self.label2idx: Dict[str, int] = {n: i for i, n in enumerate(vertices)}
        self.idx2label: Dict[int, str] = {i: k for k, i in self.label2idx.items()}

    def __getitem__(self, key: Union[str, int]) -> Union[int, str]:
        """Get either the index or vertex label based on the other one.

        :param key: The vertex label or index
        :returns: the vertex index of the label is given or the vertex label if index is given.
        """
        if isinstance(key, int):
            return self.idx2label[key]
        return self.label2idx[key]

    def __contains__(self, key: Union[str, int]) -> bool:
        """Check if the graph has a vertex labeled key.

        :param key: The vertex label or index you are asking about
        :returns: True if the vertex exists, False otherwise
        """
        if isinstance(key, int):
            return key in self.idx2label
        return key in self.label2idx

    def add_vertex(self, label: Optional[str]) -> str:
        """Add a vertex to the graph.

        :param label: The label to give the new vertex.
        :returns: The vertex label
        """
        raise NotImplementedError

    def _add_vertex(self, label: Optional[str]) -> str:
        """Add a vertex to the label2idx with a given label or a new one.

        :param label: The label for the new vertex
        :returns: The label for the new vertex
        """
        if label is None:
            label = str(len(self.label2idx))
        if label in self.label2idx:
            raise ValueError(f"Node labels must be unique, label {label} is already in use.")
        idx = len(self.label2idx)
        self.label2idx[label] = idx
        self.idx2label[idx] = label
        return idx

    def add_edge(self, source: Union[str, int], target: Union[str, int], weight: float = 1.0) -> None:
        """Add an edge to the graph.

        :param source: The vertex label or index of the edge source
        :param target: The vertex label or index of the edge target
        :param weight: The weight to put on the edge
        """
        raise NotImplementedError

    @property
    def density(self) -> float:
        """Get the density of the graph.

        The density of a graph is the ratio of edges that the graph has to the number
        it could possibly have, this is bounded by 0 and 1.

        ```math
            D = \frac{|E|}{|V|(|V| - 1)}
        ```
        """
        return self.edge_count / (self.vertex_count * (self.vertex_count - 1))

    @property
    def edge_count(self) -> int:
        """The number of edges in the graph."""
        raise NotImplementedError

    @property
    def vertex_count(self) -> int:
        """The number of vertices in the graph."""
        raise NotImplementedError

    def __str__(self) -> str:
        """A summary of the graph.

        Graph summary includes the number of vertices and edges as well as the
        density of the graph.
        """
        return f"G(V={self.vertex_count}, E={self.edge_count}, D={self.density})"

    def print_graph(self, label_lengths: Optional[int] = None) -> None:
        """Print the graph is a human readable way.

        :param label_length: A cut-off on the length of a single label while printing.
        """
        raise NotImplementedError

    def to_dot(self, directed: bool = False, label_length: Optional[int] = None) -> str:
        """Get a dot representation of the graph.

        The dot graph includes vertex labels and edge weights.

        :param directed: Should the dot representation be directed of not. Most graphs created
            in the package are directed but have the same weight in either direction so we
            can collapse the graph into an undirected weighted graph for cleaner plotting.
            Note: Collapsing this doesn't check that the weights in each direction are the same,
            it just plots a single edge.
        :param label_length: A cut-off on the length allowed for a single label in the printing.

        :returns: The representations of the graph as a dot string.
        """
        raise NotImplementedError


class AdjacencyList(Graph):
    def __init__(self, vertices: Dict[str, int]):
        """A directed simple graph represented as an AdjacencyList.

        Note:
            This is slightly different than a true adjacency list, it is a list of Vertex
            object rather than a list of lists. The Vertex objects help things like labels
            and dicts of the edges. This gives faster lookup for specific edges compared
            to a normal adjacency list.

        :param vertices: A mapping of vertex labels to integer indices or a list of vertex labels.
            If the latter then indices are assigned in order
        """
        super().__init__(vertices)
        self._vertices: List[Vertex] = [Vertex(k) for k in self.label2idx]

    @property
    def vertices(self) -> List[Vertex]:
        """The vertices in this graph."""
        return self._vertices

    def add_vertex(self, label: Optional[str]) -> str:
        """Add a vertex to the graph.

        :param label: The label to give the new vertex.
        :returns: The vertex label
        """
        label = self._add_vertex(label)
        self.vertices.insert(self.label2idx, Vertex(label))
        return label

    def add_edge(self, source: Union[str, int], target: Union[str, int], weight: float = 1.0) -> None:
        """Add an edge to the graph.

        :param source: The vertex label or index of the edge source
        :param target: The vertex label or index of the edge target
        :param weight: The weight to put on the edge
        """
        source_idx = source if isinstance(source, int) else self[source]
        target_idx = target if isinstance(target, int) else self[target]
        source_vertex = self.vertices[source_idx]
        target_vertex = self.vertices[target_idx]
        source_vertex.edges_out[target_idx] = weight
        target_vertex.edges_in[source_idx] = weight

    @property
    def vertex_count(self) -> int:
        """The number of vertices in the graph."""
        return len(self.vertices)

    @property
    def edge_count(self) -> int:
        """The number of edges in the graph."""
        return sum(v.degree_out for v in self.vertices)

    def print_graph(self, label_length: Optional[int] = None) -> None:
        """Print the graph is a human readable way.

        :param label_length: A cut-off on the length of a single label while printing.
        """
        print(str(self))
        for v in self.vertices:
            print(f"\tVertex {self[v.value]}: {v.value[:label_length]}")
            print(f"\t\tOutbound:")
            for idx, weight in v.edges_out.items():
                print(f"\t\t\t{self[v.value]} -> {idx}: {weight}")
            print(f"\t\tInbound:")
            for idx, weight in v.edges_in.items():
                print(f"\t\t\t{self[v.value]} <- {idx}: {weight}")

    def to_dot(self, directed: bool = False, label_length: Optional[int] = None) -> str:
        """Get a dot representation of the graph.

        The dot graph includes vertex labels and edge weights.

        :param directed: Should the dot representation be directed of not. Most graphs created
            in the package are directed but have the same weight in either direction so we
            can collapse the graph into an undirected weighted graph for cleaner plotting.
            Note: Collapsing this doesn't check that the weights in each direction are the same,
            it just plots a single edge.
        :param label_length: A cut-off on the length allowed for a single label in the printing.

        :returns: The representations of the graph as a dot string.
        """
        if directed:
            return self._to_directed_dot(label_length)
        return self._to_undirected_dot(label_length)

    def _to_directed_dot(self, label_length: Optional[int] = None) -> str:
        """Get a dot representation of the graph as a directed graph.

        :param label_length: A cut-off on the length allowed for a single label in the printing.

        :returns: The representations of the graph as a dot string.
        """
        dot = ["digraph G {"]
        for v in self.vertices:
            dot.append(f'\t{self[v.value]} [label="{v.value[:label_length]}"];')
            for idx, weight in v.edges_out.items():
                dot.append(f'\t{self[v.value]} -> {idx} [label="{weight}"];')
        dot.append("}")
        return "\n".join(dot)

    def _to_undirected_dot(self, label_length: Optional[int] = None) -> str:
        """Get a dot representation of the graph as a undirected graph.

        Note:
            This doesn't check that graph edges can actually be collapsed into a single edge.

        :param label_length: A cut-off on the length allowed for a single label in the printing.

        :returns: The representations of the graph as a dot string.
        """
        dot = ["graph G {"]
        edges = set()
        for v in self.vertices:
            dot.append(f'\t{self[v.value]} [label="{v.value[:label_length]}"];')
            for idx, weight in v.edges_out.items():
                if (self[v.value], idx) in edges or (idx, self[v.value]) in edges:
                    continue
                dot.append(f'\t{self[v.value]} -- {idx} [label="{weight}"];')
                edges.add((self[v.value], idx))
        dot.append("}")
        return "\n".join(dot)


class AdjacencyMatrix(Graph):
    def __init__(self, vertices: Dict[str, int]):
        """A graph with edges represented as an adjacency matrix.

        Note:
            The format of the adjacency matrix is src is the rows, tgt is the columns, i.e.:
                adjacency_matrix[src, tgt] = weight_src_to_tgt

        :param vertices: A mapping of vertex labels to integer indices or a list of vertex labels.
            If the latter then indices are assigned in order
        """
        super().__init__(vertices)
        self._adjacency_matrix = np.zeros((len(vertices), len(vertices)))

    @property
    def adjacency_matrix(self) -> np.ndarray:
        """The adjacency matrix of graph edges.

        Note:
            adjacency_matrix[x, y] is the edge starting at x and ending at y.
        """
        return self._adjacency_matrix

    def add_vertex(self, label: Optional[str]) -> str:
        """Add a vertex to the graph.

        :param label: The label to give the new vertex.
        :returns: The vertex label
        """
        label = self._add_vertex(label)
        adj = np.zeros((len(self.label2idx), len(self.label2idx)))
        adj[: self._adjacency_matrix.shape[0], : self._adjacency_matrix.shape[1]] = self._adjacency_matrix
        self._adjacency_matrix = adj
        return label

    def add_edge(self, source: Union[str, int], target: Union[str, int], weight: float = 1.0) -> None:
        """Add an edge to the graph.

        :param source: The vertex label or index of the edge source
        :param target: The vertex label or index of the edge target
        :param weight: The weight to put on the edge
        """
        source_idx = source if isinstance(source, int) else self[source]
        target_idx = target if isinstance(target, int) else self[target]
        self.adjacency_matrix[source_idx, target_idx] = weight

    @property
    def vertex_count(self) -> int:
        """The number of vertices in the graph."""
        return self.adjacency_matrix.shape[0]

    @property
    def edge_count(self) -> int:
        """The number of edges in the graph."""
        return np.sum(self.adjacency_matrix != 0)

    def print_graph(self, label_length: Optional[int] = None) -> None:
        """Print the graph is a human readable way.

        :param label_length: A cut-off on the length of a single label while printing.
        """
        print(str(self))
        for idx, label in self.idx2label.items():
            print(f"\tVertex {idx}: {label[:label_length]}")
            print(f"\t\tOutbound:")
            for i, weight in enumerate(self.adjacency_matrix[idx, :]):
                if weight == 0.0:
                    continue
                print(f"\t\t\t{idx} -> {i}: {weight}")
            print(f"\t\tInbound:")
            for i, weight in enumerate(self.adjacency_matrix[:, idx]):
                if weight == 0.0:
                    continue
                print(f"\t\t\t{idx} <- {i}: {weight}")

    def to_dot(self, directed: bool = False, label_length: Optional[int] = None) -> str:
        """Get a dot representation of the graph.

        The dot graph includes vertex labels and edge weights.

        :param directed: Should the dot representation be directed of not. Most graphs created
            in the package are directed but have the same weight in either direction so we
            can collapse the graph into an undirected weighted graph for cleaner plotting.
            Note: Collapsing this doesn't check that the weights in each direction are the same,
            it just plots a single edge.
        :param label_length: A cut-off on the length allowed for a single label in the printing.

        :returns: The representations of the graph as a dot string.
        """
        if directed:
            return self._to_directed_dot(label_length)
        return self._to_undirected_dot(label_length)

    def _to_directed_dot(self, label_length: Optional[int] = None) -> str:
        """Get a dot representation of the graph as a directed graph.

        :param label_length: A cut-off on the length allowed for a single label in the printing.

        :returns: The representations of the graph as a dot string.
        """
        dot = ["digraph G {"]
        for idx, label in self.idx2label.items():
            dot.append(f'\t{idx} [label="{label[:label_length]}"];')
            for i, weight in enumerate(self.adjacency_matrix[idx, :]):
                if weight == 0.0:
                    continue
                dot.append(f'\t{idx} -> {i} [label="{weight}"];')
        dot.append("}")
        return "\n".join(dot)

    def _to_undirected_dot(self, label_length: Optional[int] = None) -> str:
        """Get a dot representation of the graph as a undirected graph.

        Note:
            This doesn't check that graph edges can actually be collapsed into a single edge.

        :param label_length: A cut-off on the length allowed for a single label in the printing.

        :returns: The representations of the graph as a dot string.
        """
        dot = ["graph G {"]
        for idx, label in self.idx2label.items():
            dot.append(f'\t{idx} [label="{label[:label_length]}"];')
            for i, weight in enumerate(self.adjacency_matrix[idx, :]):
                if weight == 0:
                    continue
                if i >= idx:
                    continue
                dot.append(f'\t{idx} -- {i} [label="{weight}"];')
        dot.append("}")
        return "\n".join(dot)


def keyword_graph(
    tokens: List[Dict[str, str]],
    winsz: int = 2,
    sim: Callable[..., float] = lambda x, y, **kwargs: 1,
    norm: Callable[[str], str] = norm_token,
    filt: Callable[[Dict], Dict] = filter_pos,
    GraphType: Type[Graph] = AdjacencyMatrix,
) -> Tuple[Graph, Dict[str, List[int]]]:
    """Generate a keyword graph where edges are restricted to a window.

    Note:
        This also generates a dict mapping normalized vertex labels to their offsets in the original
        data. This can be used to run text rank on normalized data but return the original strings.
        It also allows the keywords post processing of combining contiguous keywords.

    :param tokens: The text to find keywords in, can be dict that have the keys `surface` and `pos` or
        just a string. If just a string then filtering of tokens is not done.
    :param winsz: The size of a window around each word that it can connect to. This window is calculated
        in the non-filtered token list.
    :param sim: A callable that returns the similarity between two vertices, used to set the weight of the edge.
        The callable should have a signature like:
            sim(
                normed_s1,
                normed_s2,
                context=context,
                raw_s1=raw_s1,
                raw_s2=raw_s2,
                raw_context=raw_context,
                s1_idx=s1_idx,
                s2_idx=s2_idx,
            ) -> float:
        Where normed_s1/2 is the normalized string of two keywords, context is all of the normalized tokens,
        raw_s1/2/context are the prenormalized versions and s1/2_idx are the indices of the keywords in the
        original sentence. This should allow complex similarity functions. The context is the whole list of
        tokens, not just the window around one.
    :param norm: A function the returns a normalized version of the input string. Default implementation
        lowercases string and removes non alpha-numeric characters.
        This is used to unify similar vertices, i.e. `Hurricane` and `hurricane` should be the same vertex
        and will be with normalization.
    :param filt: A function that will filter tokens based on pos tags if the inputs are Dicts.
    :param GraphType: The Graph class to use.

    :returns: The constructed graph and offsets mapping normalized vertex labels to their place in the original text.
    """
    # Unpack `t` in the dict (which adds all keys from t) and over-write the `surface` key
    tokens = [{**t, "normed": norm(t["surface"])} for t in tokens]

    valid_tokens = []
    offsets = defaultdict(list)
    for i, token in enumerate(tokens):
        if filt(token):
            surface = token["normed"]
            offsets[surface].append(i)
            valid_tokens.append(surface)

    vocab = build_vocab(valid_tokens)
    graph = GraphType(vocab)

    surfaces = [t["surface"] for t in tokens]
    normed_surfaces = [t["normed"] for t in tokens]

    for i, source in enumerate(tokens):
        source_surf = source["normed"]
        if source_surf not in graph:
            continue
        source_idx = graph[source_surf]
        min_ = max(0, i - winsz)
        max_ = min(len(surfaces) - 1, i + winsz)
        for j in range(min_, max_):
            if i == j:
                continue
            target = tokens[j]
            target_surf = target["normed"]
            # Don't make an edge to an un-included word or a self loop
            if target_surf not in graph or target_surf == source_surf:
                continue
            target_idx = graph[target_surf]
            graph.add_edge(
                source_idx,
                target_idx,
                sim(
                    source_surf,
                    target_surf,
                    context=normed_surfaces,
                    raw_s1=source["surface"],
                    raw_s2=target["surface"],
                    raw_context=surfaces,
                    s1_idx=source_idx,
                    s2_idx=target_idx,
                ),
            )
            graph.add_edge(
                target_idx,
                source_idx,
                sim(
                    target_surf,
                    source_surf,
                    context=normed_surfaces,
                    raw_s1=target["surface"],
                    raw_s2=source["surface"],
                    raw_context=surfaces,
                    s1_idx=target_idx,
                    s2_idx=source_idx,
                ),
            )
    return graph, offsets


def sentence_graph(
    sentences: List[str],
    sim: Callable[..., float] = overlap,
    norm: Callable[[str], str] = norm_sentence,
    GraphType: Type[Graph] = AdjacencyMatrix,
) -> Tuple[Graph, Dict[str, List[int]]]:
    """Generate a fully connected graph with edges between all sentences.

    Note:
        This also generates a dict mapping normalized vertex labels to their offsets in the original
        data. This can be used to run text rank on normalized data but return the original strings.
        You can also sort the output by offsets to make it maybe more readable?

    :param sentences: The sentences to summarize.
    :param sim: A callable that returns the similarity between two vertices, used to set the weight of the edge.
        The callable should have a signature like:
            sim(
                normed_s1,
                normed_s2,
                raw_s1=raw_s1,
                raw_s2=raw_s2,
                s1_idx=s1_idx,
                s2_idx=s2_idx,
            ) -> float:
        Where normed_s1/2 is the normalized strings of the two sentences, raw_s1/2 is the version of the sentence
        before getting normalized and s1/2_idx is the index of the sentences in the token list. This should
        facilitate both simple and complex similarity functions and also experiments that the actual flow of text
        to determine connections.
    :param norm: A function the returns a normalized version of the input sentence. Default implementation lowercases
        string and removes non alpha-numeric characters.
        This is used so simple similarity functions like the set overlap in the paper work well.
    :param GraphType: The Graph class to use.

    :returns: The constructed graph and offsets mapping normalized vertex labels to their place in the original text.
    """
    offsets = defaultdict(list)
    normed = [norm(sentence) for sentence in sentences]
    for i, norm in enumerate(normed):
        offsets[norm].append(i)

    vocab = build_vocab(normed)
    graph = GraphType(vocab)

    for (i, src), (j, tgt) in combinations(enumerate(normed), 2):
        graph.add_edge(src, tgt, sim(src, tgt, raw_s1=sentences[i], raw_s2=sentences[j], s1_idx=i, s2_idx=j))
        graph.add_edge(tgt, src, sim(tgt, src, raw_s1=sentences[j], raw_s2=sentences[i], s1_idx=j, s2_idx=i))

    return graph, offsets



import json
import argparse
from text_rank import DEMO_LOC
from text_rank.graph import sentence_graph, keyword_graph, AdjacencyList
from text_rank.text_rank import text_rank
from text_rank.keywords import keywords
from text_rank.summarize import summarize


def main():
    parser = argparse.ArgumentParser("Text Rank demo")
    parser.add_argument("--sents", "-s", type=int, default=3)
    parser.add_argument("--words", "-w", type=int, default=5)
    parser.add_argument("--iters", "-i", type=int, default=20)
    parser.add_argument("--convergence", "-c", type=float, default=0.0001)
    parser.add_argument("--verbose", "-v", action="store_true")
    parser.add_argument("--compare", action="store_true")
    args = parser.parse_args()

    if args.compare:
        sentences = json.load(open(DEMO_LOC / "automatic-summarization-sents.json"))

        G = sentence_graph(sentences)
        sents = text_rank(G, convergence=args.convergence, niter=args.iters, quiet=not args.verbose)
        print("Adjacency Matrix based Text Rank for extractive summarization")
        for i, x in enumerate(sents[: args.sents]):
            print(f" {i + 1}. {x[0]}")

        G = sentence_graph(sentences, GraphType=AdjacencyList)
        sents = text_rank(G, convergence=args.convergence, niter=args.iters, quiet=not args.verbose)
        print()
        print("Adjacency List based Text Rank for extractive summarization")
        for i, x in enumerate(sents[: args.sents]):
            print(f" {i + 1}. {x[0]}")

        tokens = json.load(open(DEMO_LOC / "automatic-summarization-tokens.json"))

        G = keyword_graph(tokens)
        kws = text_rank(G, convergence=args.convergence, niter=args.iters, quiet=not args.verbose)
        print()
        print("Adjacency Matrix based Text Rank for key-word extraction")
        for i, x in enumerate(kws[: args.words]):
            print(f" {i + 1}. {x[0]}")

        G = keyword_graph(tokens, GraphType=AdjacencyList)
        kws = text_rank(G, convergence=args.convergence, niter=args.iters, quiet=not args.verbose)
        print()
        print("Adjacency List based Text Rank for key-word extraction")
        for i, x in enumerate(kws[: args.words]):
            print(f" {i + 1}. {x[0]}")

    tokens = json.load(open(DEMO_LOC / "paper-example-keywords.json"))

    kws = keywords(tokens)
    print()
    print("Keyword Extraction from paper")
    for i, kw in enumerate(kws):
        print(f" {i + 1}. {kw}")

    sentences = json.load(open(DEMO_LOC / "paper-example-summarize.json"))

    sentences = summarize(sentences)
    print()
    print("Extractive Summarization from the paper")
    for i, sent in enumerate(sentences):
        print(f" {i + 1}. {sent}")


if __name__ == "__main__":
    main()
